#ifndef COMSERVATORY_COMSERVATORY_HPP
#define COMSERVATORY_COMSERVATORY_HPP

/**
 * @file comservatory.hpp
 *
 * @brief Entrypoint header for the **comservatory** library.
 */

/**
 * @brief Contains all **comservatory** functions and classes.
 * @namespace comservatory
 */
namespace comservatory {}

#include "ReadCsv.hpp"

#endif
